import pygame

from pygame.locals import *

from copy import deepcopy
import pickle

import hud
from objects import *

FRICTION = 0.14
GRAVITY = 0.31

class Core:
    def __init__(self,screen):
        self.screen = screen
        self.clock = pygame.time.Clock()
        self.states = {
            "game":GameState(self),
            }
        self.c_state = "game"
    def run(self):
        while True:
            self.states[self.c_state].run()

class GameState:
    def __init__(self,core):
        self.screen = core.screen
        self.clock = core.clock
        self.core = core
        self.font = pygame.font.Font(None,15)
        self.player = Player(self.screen,(40,0))
        self.area = None
        self.bg = None
        self.level = self.load_level("tutorial")
        self.checkpoint = self.check()
        self.hud = hud.HUD(self.screen,self.player)
        self.offset = None
        self.setoff()
        self.frames = 0
        screen = self.screen
        self.mask = 800
        self.imask = pygame.image.load("data/images/shield.png").convert_alpha()
        self.complete = False
        self.timer = 0
    def draw(self):
        self.screen.blit(self.bg,(0,0))
        px,py = self.player.pos
        gx,gy = self.player.gpos = ((px+20)//40,(py+20)//40)
        dlist = []
        tiles = self.level["areas"][self.area]["map"]
        items = self.level["areas"][self.area]["items"]
        enemies = self.level["areas"][self.area]["enemies"]
        for y in range(17):
            ay = (gy-8)+y
            for x in range(22):
                ax = (gx-11)+x
                dlist.append((ax,ay))
        for t in dlist:
            if tiles.has_key(t):
                tiles[t].draw(self.offset)
        for e in enemies:
            e.draw(self.offset,self.frames)
        self.player.draw(self.offset,self.frames)
        for i in items:
            i.draw(self.offset)
        pygame.draw.rect(self.screen,(0,0,0),(0,0,800,50))
        pygame.draw.rect(self.screen,(0,0,0),(0,550,800,50))
        self.hud.draw()
        y = 0
        for l in self.level["areas"][self.area]["text"]:
            self.screen.blit(self.font.render(l,True,(255,255,255)),(160,555+(y*14)))
            y += 1
        if self.mask != 800:
            self.screen.blit(self.imask,(self.mask,0))
    def load_level(self,level):
        gf = self.player.goldfeathers
        ge = self.player.eggs
        self.player = Player(self.screen,(40,0))
        self.player.goldfeathers = gf
        self.player.eggs = ge
        self.player.hp = ge*2
        self.hud = hud.HUD(self.screen,self.player)
        f = open("data/levels/%s.pickle"% level,"Ur")
        items = {"feather":Feather,
                 "gold":GoldFeather,
                 "egg":Egg,
                 "goldegg":GoldEgg,}
        enemies = {"Shroom":Shroom,
                   "Hoppin":HoppinShroom,
                   "Metal":MetalShroom,
                   "Metal2":Metal2Shroom,
                   "Wisp":Wisp,
                   "Bounce":Bouncepad,
                   "Blue":Bluepad,
                   "Orange":Orangepad,
                   "Lava":Lava}
        data = pickle.load(f)
        areas = {}
        for a in data["areas"]:
            areas[a] = {}
            areas[a]["name"] = data["areas"][a]["name"]
            areas[a]["items"] = []
            areas[a]["text"] = data["areas"][a]["text"]
            areas[a]["enemies"] = []
            areas[a]["gates"] = data["areas"][a]["gates"]
            for i in data["areas"][a]["items"]:
                areas[a]["items"].append(items[i[0]](self.screen,i[1]))
            for e in data["areas"][a]["enemies"]:
                areas[a]["enemies"].append(enemies[e[0]](self.screen,e[1],self.player))
            tiles = {}
            cx,cy = (0,0)
            for y in data["areas"][a]["map"]:
                for x in y:
                    if x in data["legend"]:
                        tiles[(cx,cy)] = Tile(self.screen,(cx*40,cy*40),data["legend"][x][1],data["legend"][x][0])
                    cx += 1
                cy += 1
                cx = 0
            areas[a]["map"] = tiles
        data["areas"] = areas
        self.player.pos = data["startpos"]
        self.player.rect.topleft = self.player.pos
        self.player.max_feathers = data["feathers"]
        self.setoff()
        self.area = data["startzone"]
        self.bg = pygame.image.load("data/images/bg-%s.png"%data["background"]).convert_alpha()
        pygame.mixer.music.load("data/sound/%s.ogg"%data["music"])
        pygame.mixer.music.play()
        return data
    def check(self):
        p = self.player
        cp = {
            "pos":p.pos,
            "level":self.level}
        return cp
    def setoff(self):
        x,y = self.player.pos
        self.offset = (380-x,280-y)
    def compass(self,entity):
        adj = {
            "n":None,
            "s":None,
            "e":None,
            "w":None,
            "ne":None,
            "nw":None,
            "se":None,
            "sw":None,
            }
        gx,gy = entity.gpos
        tiles = self.level["areas"][self.area]["map"]
        if tiles.has_key((gx,gy-1)):
            adj["n"] = (gx,gy-1)
        if tiles.has_key((gx,gy+1)):
            adj["s"] = (gx,gy+1)
        if tiles.has_key((gx+1,gy)):
            adj["e"] = (gx+1,gy)
        if tiles.has_key((gx-1,gy)):
            adj["w"] = (gx-1,gy)
        if tiles.has_key((gx+1,gy-1)):
            adj["ne"] = (gx+1,gy-1)
        if tiles.has_key((gx+1,gy+1)):
            adj["se"] = (gx+1,gy+1)
        if tiles.has_key((gx-1,gy-1)):
            adj["nw"] = (gx-1,gy-1)
        if tiles.has_key((gx-1,gy+1)):
            adj["sw"] = (gx-1,gy+1)
        return adj
    def collision(self,entity,player=False):
        e = entity
        ex,ey = e.rect.center # Lol, erect
        adj = self.compass(e)
        ghost = pygame.rect.Rect(0,0,40,40)
        ghost.center = e.rect.center
        ghost.move_ip(e.dx,e.dy)
        gx,gy = ghost.center
        tiles = self.level["areas"][self.area]["map"]
        items = self.level["areas"][self.area]["items"]
        # Check collision in each direction.
        if e.dy > 0:
            for k in ["s",]:
                if adj[k] != None and tiles.has_key(adj[k]):
                    t = tiles[adj[k]]
                    if ghost.colliderect(t.rect) and ghost.bottom >= t.rect.top:
                        ghost.bottom = t.rect.top-1
                        e.state = "supported"
                        e.dy = 0
        elif e.dy < 0:
            for k in ["n",]:
                if adj[k] != None and tiles.has_key(adj[k]):
                    t = tiles[adj[k]]
                    if ghost.colliderect(t.rect) and ghost.top <= t.rect.bottom:
                        ghost.top = t.rect.bottom+1
                        e.state = "falling"
                        e.dy = 0
        if e.dx < 0:
            m = ["w"]
            if e.state == "falling":
                m.append("sw")
            elif e.state == "jumping":
                m.append("nw")
            for k in m:
                if adj[k] != None and tiles.has_key(adj[k]):
                    t = tiles[adj[k]]
                    if ghost.colliderect(t.rect) and ghost.centerx <= t.rect.right:
                        ghost.centerx = t.rect.right+1
                        e.dx = 0                        
        elif e.dx > 0:
            m = ["e"]
            if e.state == "falling":
                m.append("se")
            elif e.state == "jumping":
                m.append("ne")
            for k in m:
                if adj[k] != None and tiles.has_key(adj[k]):
                    t = tiles[adj[k]]
                    if ghost.colliderect(t.rect) and ghost.centerx >= t.rect.left:
                        ghost.centerx = t.rect.left-1
                        e.dx = 0
        if e.state == "supported":
            supported = False
            for k in ["s",]:
                if adj[k] != None and tiles.has_key(adj[k]):
                    t = tiles[adj[k]]
                    if ghost.bottom == t.rect.top-1:
                        supported = True
            if not supported:
                e.state = "falling"
            else:
                e.dy = 0
        e.pos = e.rect.topleft = ghost.topleft
        e.hitbox.midbottom = e.rect.midbottom
        if player:
            for i in items:
                if e.rect.colliderect(i.rect):
                    ev = i.collide(e)
                    self.hud.update()
                    if ev == "Dead":
                        items.remove(i)
            for g in self.level["areas"][self.area]["gates"]:
                if e.rect.colliderect(g[0]) and e.feathers >= g[3]:
                    if g[1] != "END":
                        self.area = g[1]
                        e.pos = e.rect.topleft = g[2]
                        e.hitbox.midbottom = e.rect.midbottom
                        self.checkpoint = self.check()
                    else:
                        self.complete = True
                    
                
    def run(self):
        p = self.player
        while True:
            ms = self.clock.tick(60)
            self.frames += 1
            if p.state not in ["Dead","Complete"]:
                enemies = self.level["areas"][self.area]["enemies"]
                if self.mask != 800:
                    self.mask -= 15
                if self.mask < -1000:
                    self.mask = 800
                event = pygame.event.poll()
                keys = pygame.key.get_pressed()
                if keys[K_LEFT]:
                    p.moving = -1
                    p.facingleft = True
                elif keys[K_RIGHT]:
                    p.moving = 1
                    p.facingleft = False
                else:
                    p.moving = 0
                if event.type == KEYDOWN:
                    if event.key == K_UP:
                        if p.jumps > 0:
                            jumpnerf = ((p.goldfeathers+1)-p.jumps)
                            p.dy = -(6+p.feathers/30.0)*(0.97**float(jumpnerf))
                            p.jumps -= 1
                            p.state = "jumping"
                            self.hud.update()
                elif event.type == KEYUP:
                    if event.key == K_UP and p.dy < 0:
                        p.dy = -p.dy/2
                if p.moving == 0:
                    if p.dx > 0:
                        p.dx = max(0,p.dx-FRICTION)
                    if p.dx < 0:
                        p.dx = min(0,p.dx+FRICTION)
                elif p.moving == -1:
                    p.dx = max(p.dx-p.a,-3)
                elif p.moving == 1:
                    p.dx = min(p.dx+p.a,3)
                p.dy = min(p.dy+GRAVITY,p.tv)
                self.collision(self.player,player=True)
                if p.dy > 0:
                    p.state = "falling"
                if p.state == "supported":
                    p.dy = max(0.0,p.dy)
                    p.jumps = p.goldfeathers+1
                    self.hud.update()
                p.immunity = max(0,p.immunity-1)
                for x in self.level["areas"][self.area]["enemies"]:
                    x.ai("general")
                    if x.moving == 0:
                        if x.dx > 0:
                            x.dx = max(0,x.dx-FRICTION)
                        if x.dx < 0:
                            x.dx = min(0,x.dx+FRICTION)
                    elif x.moving == -1:
                        x.dx = max(x.dx-x.a,-x.maxspeed)
                    elif x.moving == 1:
                        x.dx = min(x.dx+x.a,x.maxspeed)
                    x.dy = min(x.dy+GRAVITY,x.tv)
                    self.collision(x)
                    x.immunity = max(0,x.immunity-1)
                    if x.dy > 0:
                        x.state = "falling"
                    if x.state == "supported":
                        x.dy = max(0.0,x.dy)
                        
                    if p.hitbox.colliderect(x.hitbox):
                        if p.state == "falling" and x.can_hit:
                            if x.immunity == 0:
                                x.hp -= 1
                                if x.hurtnoise != None:
                                    x.hurtnoise.play()
                                self.hud.update()
                                x.immunity = 50
                            if x.hp == 0:
                                enemies.remove(x)
                            if keys[K_UP]:
                                p.dy = -6*x.bounce
                            else:
                                p.dy = -4*x.bounce
                        elif p.immunity == 0:
                                p.hp = max(0,p.hp-x.damage)
                                p.hurtnoise.play()
                                p.immunity = 80
                                dx = p.pos[0]-x.pos[0]
                                dy = p.pos[1]-x.pos[1]
                                p.dx = dx/5
                                p.dy = dy/5
                self.setoff()
                self.draw()
                pygame.display.flip()
                if event.type == QUIT:
                    raise SystemExit()
                                
                if p.hp == 0 or p.pos[1] > 3000:
                    p.state = "Dead"
                    pygame.mixer.Sound("data/sound/chicken-fade.ogg").play()
                    self.mask = 800
                if self.complete:
                    p.state = "Complete"
                    self.mask = 800
            else:                
                if p.state in ("Dead","Complete") and self.mask > -100:
                    self.mask -= 15
                if self.mask <= -80:
                    if p.state == "Dead":
                        self.mask = -100
                        c = self.checkpoint
                        self.level = c["level"]
                        p.pos = p.rect.topleft = c["pos"]
                        p.hitbox.midbottom = p.rect.midbottom
                        p.dx = p.dy = 0
                        p.state = "falling"
                        p.moving = 0
                        p.immunity = 0
                        p.hp = p.eggs*2
                    if self.player.state == "Complete":
                        self.mask = -100
                        nl = self.level["next"]
                        self.timer += 1
                        if self.timer > 180:
                            self.complete = False
                            self.timer = 0
                            self.level = self.load_level(nl)
                            self.checkpoint = self.check()
                            self.core.run()
                event = pygame.event.poll()
                if event.type == QUIT:
                    raise SystemExit()
                self.setoff()
                self.draw()
                pygame.display.flip()


